/*
 *	(c) Copyright 1993 Paramax Systems Corp.
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <search.h>
#include <string.h>
#include <malloc.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

struct performative {
	char *name;
	handler hfunction;
	int blocking;
	};

static int perfcmp (const void *p1, const void *p2);

static struct performative *generic_perf = NULL;
static struct performative *ptable = NULL;

static int
perfcmp(const void *p1, const void *p2)
{
	return (kqml_stricmp (((struct performative *) p1)->name, 
						((struct performative *) p2)->name));
}

handler
_kqml_find_handler(char *name)
{
	struct performative p, **perf;
	handler ret_value;
  
	if (name == NULL || name[0] == '\0')
	{
		if (generic_perf)
		{
			ret_value = generic_perf->hfunction;
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("_kqml_find_handler: Returning default handler.");
		}
		else
		{
			ret_value = NULL;
		}

	}
	else
	{
		p.name = name;
		perf = (struct performative **) tfind (&p, (void **)&ptable, perfcmp);
		if (perf == NULL)
		{
			if (generic_perf)
			{
				ret_value = generic_perf->hfunction;
				_kqml_dbg_msg (KQML_DBG_VERBOSE) 
					("_kqml_find_handler: Returning default handler.");
			}
			else
			{
				ret_value = NULL;
			}
		}
		else
		{
			ret_value = (*perf)->hfunction;
		}
	}

	return ret_value;
}

/*
 * kqml_register
 *
 * Register a function to asynchronously receive and handle KQML messages.
 *
 *	1 - OK, the function is registered
 *    < 0 - error code
 */
int
kqml_register (handler hfunction, char *perfname, int blocking)
{
	struct performative *perf, **ret_value;
  
	/* Find out if a handler is already defined.
		If one is already defined. the ret_value will be non-null 
	*/
	if (perfname == NULL || perfname[0] == '\0')
	{
		perfname = ""; /* Just in case it was NULL */
		/* Default handler */
		if (generic_perf != NULL)
		{
			ret_value = &generic_perf;
		}
		else
		{
			ret_value = NULL;
		}
	}
	else
	{
		struct performative p;

		p.name = perfname;
		ret_value = (struct performative **) tfind (&p, (void **)&ptable, perfcmp);
	}	

	/* If the ret_value == NULL, allocate space and 
		store in the tree if it is not a default handler */
	if (ret_value == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Adding [%s] performative.", perfname);
		perf = malloc (sizeof (struct performative));
		if (perf == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Not enough space for [%s] performative slot.", perfname);
				return -1;
		}
		perf->name = strdup (perfname);
		if (perf->name == NULL)
		{
			free (perf);
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Not enough space for [%s] performative.", perfname);
			return -1;
		}

		if (perfname[0] == '\0')
		{
			/* default_handler */
			generic_perf = perf;
		}
		else
		{
			ret_value = (struct performative **) tsearch (perf, (void **)&ptable, perfcmp);
			if (ret_value == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("Not enough space for %s performative in tree\n", perfname);
				free (perf->name);
				free (perf);
				return -1;
			}
		}
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("Replacing [%s] performative.", perfname);
		perf = *ret_value;
	}

	perf->hfunction = hfunction;
	perf->blocking = blocking;

	return 1;
}
