/*
 * $Log: init.c,v $
 * Revision 1.6  1993/05/10  17:59:14  robin
 * Added processing to add facilitator timings to overrall timings.
 *
 * Revision 1.5  1993/04/25  07:51:26  robin
 * Stable with SAS KRIL.
 *
 * Revision 1.4  1993/01/26  22:08:49  robin
 * Fixed little bugs here and there.
 *
 * Revision 1.3  1993/01/25  20:41:06  robin
 * Modified code to always pass the application name as a command line
 * argument to the router in the execv call.
 *
 * Revision 1.2  1993/01/25  20:12:58  robin
 * Allowed for router_debug_level to be entered from the command line.
 *
 * Revision 1.1  1993/01/25  13:47:27  fritzson
 * Initial revision
 *
 *
 */


/*
 *	(c) Copyright 1993 Paramax Systems Corp.
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

#include <kqml/kqmlconf.h>

#include	<sys/time.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/param.h>
#include <sys/wait.h>

#include <netinet/in.h>
#include	<arpa/inet.h>

#include	<stdlib.h>
#include	<stdio.h>
#include <errno.h>
#include	<signal.h>
#include	<rpc/rpc.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<string.h>
#include <memory.h>

#if defined (__KQML_MACH_SUN__)
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include	<kqml/kqmllib.h>

/*
 * Exported Functions:
 *
 * kqml_initialize (application_name, ans_handler ans_lookup_func)
 *
 * Variables exported to other modules in the library:
 * kqml_agent_name - stores the string passed in to kqml_initialize
 */

/* globally saved name of application */
char *kqml_agent_name = NULL, *kqml_ans_name = NULL, 
	*kqml_protocol = NULL, *kqml_addr_info = NULL;

int _kqml_im_the_router = FALSE;
int _kqml_im_an_ans = FALSE;
int _kqml_root_ans = FALSE;

ans_handler _kqml_ans_lookup = NULL;

static int inside_handler = FALSE;
static char child_started = FALSE;	/* Tells whether router or validator 
													is alive */
static int child_pid = 0;				/* router or validator process pid */
static char *child_name;				/* name of router or validator */
static int from_child, to_child;		/* Connection to the child */
static FILE *from_child_stream, 		/* Stream connection to the child */
				*to_child_stream; 
static int timings_flag = FALSE;		/* Timings flag */		
static int logging_flag = FALSE;		/* Logging flag */

static char *get_child_path (void);
static int start_child (char *app_name, char *child_path, 
	FILE *dbg_file, int dbg_level);
static void handle_child (int sig, void *dummy1, void *dummy2);
static void handle_child_death (int sig, void *dummy1, void *dummy2);
static int restart_child ();

static int critical_section (int start_flag);
static void signal_catcher (int signal_no, void *dummy1, void *dummy2);
static void kill_child ();
static void terminate_child (void);
static void terminate_child_sun (int status, caddr_t arg);

/*
 * Router Initialization routines 
 *
 * This routine initialize the KQML router and the application's
 * connection to it.
 *
 */

/*
 * kqml_initialize
 *
 * Initialize the KQML router and a connection to it.
 *
 * Parameters:
 *		ans_flag
 *		TRUE, if this agent is an ANS. FALSE otherwise
 * 	application_name
 *	   The symbolic name of the application. This is advertised to
 *	   an ANS and is used by other routers to locate this
 *	   application.
 *	   
 * Value:
 *	1 - OK
 *    < 0 - Error code
 * 
 *
 * Description:
 *
 *    This routine starts the KQML router as a separate process and
 *    registers the application name with the router.	       
 */
int
kqml_initialize (char *app_name, ans_handler ans_lookup)
{
	char *child_path; 			/* router/validator executable */
	kqml_message *msg;
	int res, dbg_level;
	FILE *dbg_file;
	char *dbg_file_ptr, *dbg_level_ptr, *content;
	sigset_t signal_mask;
	struct sigaction act;
  
	_kqml_im_the_router = FALSE;
	_kqml_im_an_ans = (ans_lookup ? TRUE : FALSE);
	_kqml_ans_lookup = ans_lookup;
	kqml_agent_name = app_name; /* Temporarily, till the final value is set */

	dbg_level_ptr = getenv ("KQML_DBG_LEVEL");
	if (dbg_level_ptr != NULL)
	{
		dbg_level = atoi (dbg_level_ptr);
		if (dbg_level < KQML_DBG_ALWAYS)
			dbg_level = KQML_DBG_ALWAYS;
	}
	else
	{
		dbg_level = KQML_DBG_ALWAYS;
	}
	dbg_file_ptr = getenv ("KQML_DBG");
	/* Only if the KQML_DBG variable is set */
	if ( dbg_file_ptr )
	{
		char file_name[256];

		sprintf (file_name, "%s/%s.Kdbg", dbg_file_ptr, app_name);
		unlink (file_name);
		dbg_file = fopen (file_name, "w");
		if (dbg_file)
		{
			setvbuf(dbg_file, NULL, _IOLBF, 0);
		}
	}
	else
	{
		dbg_file = NULL;
	}

	_kqml_init_debugging (app_name, dbg_file, dbg_level);

	res = strlen (app_name);
	if (_kqml_check_word (app_name, res) != 0 || 
		memchr (app_name, '.', res) != NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Invalid characters in agent name. [%s]", app_name);
		return -1;
	}
		
	child_path = get_child_path ();

	sigemptyset(&signal_mask);
	sigaddset(&signal_mask, SIGIO);
	act.sa_handler = handle_child_death;
	act.sa_mask = signal_mask;

#	if defined (__KQML_MACH_SVR4__)
	act.sa_flags = SA_RESTART;
#	else
	act.sa_flags = 0;
#	endif

	if ( (int) (sigaction (SIGCHLD, &act, NULL)) <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to install SIGCHLD handler. %s.",
			sys_errlist[errno]);
		return -1;
	}
  
	/* start the router or validator*/
	if (start_child (app_name, child_path, dbg_file, dbg_level) <= -1)
	{
		return (-1);
	}

	from_child_stream = fdopen (from_child, "r");
	to_child_stream = fdopen (to_child, "w");

	if (from_child_stream == NULL || to_child_stream == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to create file streams.");
		return -1;
	}

#if 1
  sleep (1);			/* WHY? */
#endif

	if (! _kqml_init_xdr (from_child_stream, to_child_stream))
	{
		return -1;
	}

	/* Establish SIGIO handler */
	sigemptyset (&signal_mask);
	act.sa_handler = handle_child;
	act.sa_mask = signal_mask;
#	if defined (__KQML_MACH_SVR4__)
	act.sa_flags = SA_RESTART;
#	else
	act.sa_flags = 0;
#	endif

	if ( (int) (sigaction (SIGIO, &act, NULL)) < 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to install SIGIO handler. %s.", 
			sys_errlist[errno]);
		kill (child_pid, SIGTERM);
		return (-1);
	}

	/* We wont be setting FIOSETOWN and FIOASYNC on from_child.
		'cos the SIGIO facility is 4.3BSD specific.
		Instead the router will send us a SIGIO after writing 
		into the XDR Stream
	*/

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml_initialize: Waiting for %s.",
		child_name);

	/* wait for child handshake  */
	/* I don't think this is necessary. 'cos if the router dies,
		we will get a child dead signal and we will die -Chelliah */
	res = _kqml_wait_for_msg (20);
  
	/* No message from child */
	/* await returns 0 on timeout, < 0 on error and > 0 on success */
	if (res <= 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to verify %s is up.", 
			child_name);
		kill (child_pid, SIGTERM);
		return (-1);
	}

	/* read a confirmation message from the child */
	msg = _kqml_fetch_xdr_msg ();

	/* Message from child is empty */
	if (! msg)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("No handshake message from %s.\n", child_name);
		kill (child_pid, SIGTERM);
		return (-1);
	}
  
	/* Message from child is not the right one */
	if ( kqml_stricmp(kqml_get_performative (msg), KQML_HANDSHAKE) != 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Incorrect %s Version. [%s]\n", 
			child_name, kqml_get_performative (msg));
      kill (child_pid, SIGTERM);
      return (-1);
 	}
	kqml_free_msg (msg);

	/* build the reply message */
	msg = kqml_build_msg (KQML_HANDSHAKE, app_name, NULL, NULL, NULL);
	if (msg)
	{
		_kqml_send_xdr_msg (msg);
		kqml_free_msg (msg);
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Could not construct greeting message.");
		kill (child_pid, SIGTERM);
		return -1;
	}

	/* read the third message from the child */
	msg = _kqml_fetch_xdr_msg ();

	/* Message from child is empty */
	if (! msg)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("No greeting  message from %s.\n", child_name);

		/** Commented out to check if this works for duplicate registration
				-- Anupama, Feb. 6, '96. **/

/*		kill (child_pid, SIGTERM);*/
		return (-1);
	}
  
	/* Message from child is not the right one */
	if ( kqml_stricmp(kqml_get_performative (msg), KQML_HANDSHAKE) != 0 ||
		(content = kqml_get_field (":content", msg, NULL, NULL)) == NULL )
		
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Incorrect %s Version. [%s]\n", 
			child_name, kqml_get_performative (msg));
      kill (child_pid, SIGTERM);
      return (-1);
 	}

	content = strchr (content, '(');
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Illegal greeting from %s.", child_name);
		kill (child_pid, SIGTERM);
		return (-1);
	}
	content++;

	kqml_agent_name = strtok (content, " \t\n");
	if (kqml_agent_name == NULL || 
		(kqml_agent_name = strdup (kqml_agent_name)) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for agent name.");
		kill (child_pid, SIGTERM);
		return (-1);
	}

	kqml_ans_name = strtok (NULL, " \t\n");
	if (kqml_ans_name == NULL ||
		(kqml_ans_name = strdup (kqml_ans_name)) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for ans name.");
		kill (child_pid, SIGTERM);
		return (-1);
	}

	kqml_protocol = strtok (NULL, " \t\n");
	if (kqml_protocol == NULL ||
		(kqml_protocol = strdup (kqml_protocol)) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for protocol.");
		kill (child_pid, SIGTERM);
		return (-1);
	}

	kqml_addr_info = strtok (NULL, " \t\n)");
	if (kqml_addr_info == NULL ||
		(kqml_addr_info = strdup (kqml_addr_info)) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for addr_info.");
		kill (child_pid, SIGTERM);
		return (-1);
	}

	if (kqml_stricmp (kqml_ans_name, kqml_agent_name) == 0)
	{
		_kqml_root_ans = TRUE;
	}
	kqml_free_msg (msg);

	child_started = TRUE;

#if defined(__KQML_MACH_SVR4__)
	atexit (terminate_child);
#elif defined(__KQML_MACH_SUN__)
	on_exit (terminate_child_sun, NULL);
#endif

	sigemptyset (&signal_mask);
	act.sa_handler = signal_catcher;
	act.sa_mask = signal_mask;
#	if defined (__KQML_MACH_SVR4__)
	act.sa_flags = SA_RESTART;
#	else
	act.sa_flags = 0;
#	endif

	sigaction (SIGUSR2, &act, NULL);
	sigaction (SIGTERM, &act, NULL);
	sigaction (SIGSEGV, &act, NULL);
	sigaction (SIGBUS, &act, NULL);
	sigaction (SIGINT, &act, NULL);
	sigaction (SIGQUIT, &act, NULL);

	return (1);
}

int
kqml_critical_section_start ()
{
	return critical_section (TRUE);
}

int
kqml_critical_section_end ()
{
	return critical_section (FALSE);
}

static int 
critical_section (int start_flag)
{
	if (! inside_handler)
	{
		sigset_t signal_mask;
		int how;

		how = (start_flag ? SIG_BLOCK : SIG_UNBLOCK);
		sigemptyset (&signal_mask);
		sigaddset (&signal_mask, SIGIO);
		if ( (sigprocmask (how, &signal_mask, NULL)) < 0)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("critical_section: Unable to modify the SIGIO signal.");
			return FALSE;
		}
	}

	return TRUE;
}

static char *
get_child_path ()
{
	struct stat buf;
	static char local_path[MAXPATHLEN];
	char *child_path;

	/* Find out if validator is defined */
	if (getenv (KQML_VALIDATOR_FLAG))
	{
		child_name = KQML_VALIDATOR_NAME;
		child_path = getenv (KQML_VALIDATOR_ENV);
	}
	else
	{
		child_name = KQML_ROUTER_NAME;
		child_path = getenv (KQML_ROUTER_ENV);
	}

	if (child_path == NULL || stat (child_path, &buf) != 0)
	{
		child_path = getenv ("KQML_HOME");
		if (child_path)
		{
			strcpy (local_path, child_path);
			strcat (local_path, "/bin/");
			strcat (local_path, child_name);
			child_path = local_path;
		}
	}

	if (child_path == NULL || stat (child_path, &buf) != 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unable to find %s using env variables\n", child_name);
		return child_name;
	}
	else
	{
		return child_path;
	}
}

static int
start_child (char *app_name, char *child_path, FILE *dbg_file, int dbg_level)
{
	int i, fimo[2], mifo[2];
	char *child_args[6], dbg_level_str[10];
  
	/* create communication pipes */
	if (pipe (fimo)) 
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Can't make pipe 1. %s.", sys_errlist[errno]);
		return (-1);
	}

	if (pipe (mifo)) 
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Can't make pipe 2. %s.", sys_errlist[errno]);
		return (-1);
	}
  
	/* fork router process */
	child_pid = fork();
	switch (child_pid) 
	{
	case -1:
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to fork process for %s. %s.", 
			child_path, sys_errlist[errno]);
		return (-1);
		break;

	case 0:			/* new process */
		/* stdin <- input of fimo */
		if (dup2 (fimo[0], 0) == -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to dup2 input.");
			return -1;
		}
		close (fimo[1]);
		close (fimo[0]);
		/* stdout <- output of mifo */
		if (dup2 (mifo[1], 1) == -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to dup2 output.");
			return -1;
		}
		close (mifo[0]);
		close (mifo[1]);
		/* Link the log files */
		if (dbg_file)
		{
			if (dup2 (fileno (dbg_file), 2) == -1)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to dup2 error.");
				/* Not a serious error */
			}
		}

		/*  start child */
		i = 0;
		child_args[i++] = child_path;

		if (timings_flag)
		{
			child_args[i++] = "-T";
		}

		if (logging_flag)
		{
			child_args[i++] = "-L";
		}

		sprintf (dbg_level_str, "%d", dbg_level);

		child_args[i++] = "-d";
		child_args[i++] = dbg_level_str;

		/* TAKE NOTE: FACILITATOR IS SPECIAL */
		if (_kqml_im_an_ans)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("This is an ANS.");
			child_args[i++] = "-f";
		}
		child_args[i] = NULL;
		if ( (execvp (child_path, child_args)) < 0)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unable to exec %s. %s.", child_path,
				sys_errlist[errno]);
			return (-1);
		}
		break;

	default:
		close (fimo[0]);
		close (mifo[1]);
		from_child = mifo[0];
		to_child = fimo[1];
		break;
	}
	return 0;
}

static 
int restart_child ()
{
	return -1;
}

static void
signal_catcher (int signal_no, void *dummy1, void *dummy2)
{
	if (signal_no == SIGUSR2)
	{
		if (child_pid)
			kill (child_pid, SIGUSR2);
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Application recd signal %d.", 
			signal_no);
		kill_child ();
		exit (1);
	}
}

static void 
handle_child (int sig, void *dummy1, void *dummy2)
{
	kqml_message *kmsg;

	inside_handler = TRUE;
	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("Signal Handling message from %s.", child_name);

	/* await returns 0 on timeout, < 0 on error and > 0 on success */
	if (_kqml_wait_for_msg (1) <= 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("False alarm. Nothing from %s.", child_name);
	}
	else
	{
		do
		{
			kmsg = _kqml_fetch_xdr_msg ();

			if (kmsg == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_VERBOSE)
					("No KQML message. So, return already.:");
				break;
			}

			_kqml_handler (kmsg);
		}
		while (_kqml_msgs_in_queue ());
	}
	inside_handler = FALSE;
}

static void 
handle_child_death (int sig, void *dummy1, void *dummy2)
{
	int pid;

	pid = waitpid (child_pid, NULL, WNOHANG);
	if (pid)	/* the router/validator is dead, so try to restart it */
	{
		/* Only if it happens after router has been started once before */
		if (child_started && restart_child () < 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("The KQML %s has ceased to function.", child_name);
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("I now go to join it.");
			exit (1);
		}
	}
}

static void
terminate_child_sun (int status, caddr_t arg)
{
	terminate_child ();
}

static void 
terminate_child ()
{
	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("Application is exiting. Killing %s.", child_name);

	kill_child ();
}

static void 
kill_child ()
{
	struct sigaction act;
	sigset_t signal_mask;
	static int going = 1;

	if (!going) 
		return;
	going = 0;
  
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Sending SIGTERM to %s.", child_name);

	sigemptyset (&signal_mask);
	act.sa_mask = signal_mask;
#	if defined (__KQML_MACH_SVR4__)
	act.sa_flags = SA_RESTART;
#	else
	act.sa_flags = 0;
#	endif
	act.sa_handler = SIG_IGN;
	sigaction (SIGIO, &act, NULL);
	sigaction (SIGCHLD, &act, NULL);
	kill (child_pid, SIGTERM);
}

void kqml_timings_start ()
{
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Turning KQML timings ON.");
	timings_flag = TRUE;
	if (child_started)
	{
		kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(timings on)", 
			kqml_agent_name, NULL, NULL);
	}
}

void kqml_timings_stop ()
{
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Turning KQML timings OFF.");
	timings_flag = FALSE;
	if (child_started)
	{
		kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(timings off)", 
			kqml_agent_name, NULL, NULL);
	}
}

int kqml_timings ()
{
	return timings_flag;
}

void kqml_logging_start ()
{
	logging_flag = TRUE;
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Turning KQML logging ON.");
	if (child_started)
	{
		kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(logging on)", 
			kqml_agent_name, NULL, NULL);
	}
}

void kqml_logging_stop ()
{
	logging_flag = FALSE;
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Turning KQML logging OFF.");
	if (child_started)
	{
		kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(logging off)", 
			kqml_agent_name, NULL, NULL);
	}
}

int kqml_logging ()
{
	return logging_flag;
}
