/*
 * router.c
 *
 * A KQML router.
 * Richard Fritzson & Robin McEntire
 * Written: December 1992
 * Revised: October  1993
 * Revised: July/Aug 1995 by
 *             Robin McEntire, Loral Defense Systems-Eagan
 *             Chelliah Thirunavukkarasu, UMBC
 *             Anupama Potluri, UMBC
 *             R. Scott Cost, UMBC
 *

 *    (c) Copyright 1993 Loral Defense Systems-Eagan
 *
 *    This material my be reproduced by or for the U.S.
 *    Government pursuant to the copyright license under
 *    the clause at DFARS 252.227-7013   (1991).
 *

 * The KQML router is designed to be invoked by an application to handle
 * its KQML connectivity. This invocation is handled by KQML 
 * initialization routines so this program is never directly invoked by
 * a user.
 *
 * Nevertheless, for the sake of clarity, here are the options which
 * can be passed on the command line:
 *
 * -f						provided by ANS process only
 * -d dbg_level		amount of routine debugging
 * -L						Enable Logging functionality
 * -T						Enable Timings functionality
 *
 */

/*
 * The KQML router is an auxilliary process which is associated with a
 * single KQML application. It receives all incoming KQML messages for
 * the application and routes all outgoing process for it. 
 *
 * Communication between the router and the appliation is via a UNIX
 * pipe, carrying an XDR representation of a parsed KQML message. 
 *
 * A description of the representation can be found in src/lib/kqmlmsg.x.
 */

#include <kqml/kqmlconf.h>

#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <time.h>
#include <memory.h>
#include <rpc/rpc.h>
#include <stdio.h>
#include <netdb.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifdef __KQML_MACH_SGI__
#include <bstring.h>
#endif

#ifdef __KQML_MACH_SVR4__
#include <ulimit.h>
#include <sys/systeminfo.h>
#endif

#ifdef __KQML_MACH_SUN__
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

#define MAX_ROUTER_CONNECTIONS	64

#define APP_IN		0
#define APP_OUT		1
#define KQML_HOSTNAME_LEN			256

extern char	*optarg;

static kqml_message *parse_kqml (int con);
static int parse_tcp_info (int env_flag, char *param, char **host, int *port);

static int init_router (void);
static int init_app ();
static int establish (int portnum, long addrtype);
static int init_listener ();
static int init_ans ();
static int register_application (int status);
static int confirm_registration ();
static int complete_handshake ();

static int mainloop ();
static int kqml_logging_msg (kqml_message *kmsg);

static int get_active_connection (int previous_con);
static int accept_new_connection ();
static int close_nonpending_connection ();
static int verify_connection_range (int con);
static void dont_linger (int con);

static int route_msg_to_application (int con, kqml_message **k_msg);
static int send_to_application (int con, kqml_message *kmsg);
static int route_msg_from_application (int con, kqml_message **k_msg);
static int send_to_remote (int con, kqml_message *kmsg, struct timeval *beg, 
	struct timeval *end);

static int get_connection_to_remote(char *symbolic_name, char *lookup_info);
static int lookup_and_connect(char *agent_name, char *ans_name, int ans_con);
static int connect_to_agent(int env_flag, char *agent_name, char *addr_info);

static void send_error_reply (char *msg, kqml_message *old_msg);
static int handle_admin_msg (kqml_message *kmsg);

static void sigpipe_handler (int sig, void *dummy1, void *dummy2);
static void router_signal_handler (int signo, void *dummy1, void *dummy2);
static void router_exit (int ret_code);

extern int send_parsed_msg (kqml_message *msg, int con, FILE *stream);
extern int set_kmsg_time (char *kmsg_id, int timings_type, 
	struct timeval *beg_time, struct timeval *end_time);


/*								*/
/* Connection Table Routines 					*/
/*								*/

/* 
 * All active connections to other KQML routers are maintained in a
 * connection table which describes the conntection. The table is
 * indexed by the socket number associated with the connection.
 * These routines and macros implement the interface to the set of 
 * active connections with other KQML routers.
 * 
 */

/*
 * init_connections ()		- initialize the datastructure 
 * disable_connection (n)	- clear a connection entry 
 * enable_connection (n, s)	- turn on a connection
 * 				  Accessor Macros
 * connection_name(n) 		- symbolic name of remote agent
 * connection_host(n) 		- host name of remote agent
 * connection_port(n) 		- TCP/IP port of remote agent
 * connection_pending(n) 	- true if waiting for a reply on connection n
 * connection_instream(n) 	- read stream to remote agent
 * connection_outstream(n) - write stream to remote agent
 * connection_last_use_time(n)  - time of last i/o operation 
 */

/*
 * WARNING: connection_* function does not check if "n" is within
 * the con_table bounds. We can easily add it by modifying the
 * macro. But it will add additional overhead and may not be
 * necessary always. So, it is left to the programmer to check
 * it explicitly when he/she thinks it is necessary.
 */
#define connection_valid(n)			((n) >= 0 && n < MAX_ROUTER_CONNECTIONS)
#define connection_name(n)				con_table[(n)].name
#define connection_host(n) 			con_table[(n)].host
#define connection_port(n) 			con_table[(n)].port
#define connection_pending(n) 		con_table[(n)].pending_flag
#define connection_instream(n) 		con_table[(n)].instream
#define connection_outstream(n) 		con_table[(n)].outstream
#define connection_create_time(n) 	con_table[(n)].create_time
#define connection_last_use_time(n)	con_table[(n)].last_use_time

/* Not Public */
static fd_set servers;
static struct centry 
{
	char *name;
	char *host;
	long port;
	int pending_flag;
	FILE *instream;
	FILE *outstream;
	long create_time;			/* If 0, entry is unused */
	long last_use_time;
} con_table[MAX_ROUTER_CONNECTIONS];


/*
 * We use a creation_time of zero to signify an unused connection entry.
 *
 */

static void 
disable_connection (int con)
{
	connection_create_time (con) = 
		connection_last_use_time (con) = 0;
	connection_pending (con) = FALSE;
	connection_port (con) = 0;

	if (connection_name (con) != NULL)
	{
		free (connection_name (con));
		connection_name (con) = NULL;
	}
	if (connection_host (con) != NULL)
	{
		free (connection_host (con));
		connection_host (con) = NULL;
	}
	if (connection_instream (con) != NULL)
	{
		fclose (connection_instream (con));
		connection_instream (con) = NULL;
	}
	if (connection_outstream (con) != NULL)
	{
		fclose (connection_outstream (con));
		connection_outstream (con) = NULL;
	}

	FD_CLR (con, &servers);
}

/*
 * if streams is true, then open i/o streams to this con
 */

static void
enable_connection (int con, int streams)
{
	connection_create_time (con) = time (NULL);
	connection_last_use_time (con) = connection_create_time (con);
	connection_pending (con) = FALSE;
	FD_SET (con, &servers);

	if (streams)
	{
		connection_instream (con) = fdopen (con, "r");
		connection_outstream (con) = fdopen (con, "w");
	}
	else
	{
		connection_instream (con) = NULL;
		connection_outstream (con) = NULL;
	}
}

static void
init_connections ()
{
	int i;
  
	for (i = 0; i < MAX_ROUTER_CONNECTIONS; i++)
	{
		disable_connection (i);
	}
	FD_ZERO (&servers);
}

static void
dump_connection_table ()
{
	int i;

	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("*** router connection table start ***");
	for (i = 0; i < MAX_ROUTER_CONNECTIONS; i++)
	{
		if (connection_create_time (i) != 0)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("%d %s %s %d %s %x %x %d %d", i,
				(connection_name (i) ? connection_name (i) : "unknown"),
				(connection_host (i) ? connection_host (i) : "unknown"),
				connection_port (i), 
				(connection_pending (i) ? "Pending    " : "Not Pending"),
				connection_instream (i), connection_outstream (i),
				connection_create_time (i), connection_last_use_time (i));
		}
	}
	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("*** router connection table end   ***");
}


static int agent_pid = 0;
static int ans_connection = -1;			/* fd of ans connection */
static int logger_connection = -1; 		/* fd of KQML-LOGGER connection */
static int kqml_listener;					/* listener fd */

int
main (int argc, char *argv[])
{
	int c;
	int dbg_level = KQML_DBG_ALWAYS;		/* debugging level */
	FILE *dbg_file = NULL;	      		/* file for writing debug msgs to */
	char *ans_addr;

	agent_pid = getppid ();
	_kqml_im_the_router = TRUE;

	/* Process command line arguments */
	while ((c = getopt (argc, argv, "fd:LT")) != -1)
	switch (c)
	{
	case 'f':
		_kqml_im_an_ans = TRUE;
		break;
	case 'd':
		dbg_level = atoi (optarg);
		if (dbg_level < KQML_DBG_ALWAYS)
			dbg_level = KQML_DBG_ALWAYS;
		break;
	case 'L':
		kqml_logging_start ();
		break;
	case 'T':
		kqml_timings_start ();
		break;
	case '?':
		fprintf (stderr,
			"Router was passed unknown argument. Exiting.\n");
		exit (1);
		break;
	}

	_kqml_init_debugging (KQML_ROUTER_NAME, dbg_file, dbg_level);
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Initialized.");


	/* initialize */
	if (init_router () <= -1)
	{
		exit (1);
	}

	if (init_app () <= -1)
	{
		exit (1);
	}

	if (init_listener () <= -1)	/* listener */
	{
		exit (1);
	}

	if (_kqml_im_an_ans)
	{
		ans_addr = getenv ("KQML_PARENT_ANS");
	}
	else
	{
		ans_addr = getenv ("KQML_ANS");

		if (ans_addr == NULL)
		{
			/* We are not an ANS and we are not connected to any ANS */
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("KQML_ANS not defined.");
			exit (1);
		}
	}

	if (ans_addr)
	{
		if (init_ans (ans_addr) <= -1)
		{
			exit (1);
		}

		/* We are connected to a parent ans */
		if (register_application (TRUE) <= -1)
		{
			exit (1);
		}
		
		if (! confirm_registration ())
		{
			exit (1);
		}

		connection_name (ans_connection) = strdup (kqml_ans_name);
	}
	else
	{
		/* We do not have a parent ANS */

		_kqml_root_ans = TRUE;
		kqml_agent_name = strdup (kqml_agent_name);
		if (kqml_agent_name == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for agent name.");
			exit (1);
		}
		kqml_ans_name = kqml_agent_name;
	}

	if (! complete_handshake ())
	{
		exit (1);
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Starting router now.");
	mainloop ();		/* run */

	router_exit (0);		/* quit */

	return 0;		/* NOT REACHED */
}

/*
 * By the time you call this, you have a stream open to the application, 
 * a stream open to the ans and a listener on the kqml port.
 *
 * Perform a select on all open connections and handle any inputs. 
 *
 */
static int
mainloop ()
{
	int con = APP_IN;
	kqml_message *kmsg;
  
	while (1)
	{
		kmsg = NULL;
		con = get_active_connection (con);
		
		if (con <= -1)
		{
			_kqml_dbg_msg (KQML_DBG_WARNING) ("get_active_connection failed.");
		}
		else if (APP_IN == con)	/* message from the application */
		{
			if (route_msg_from_application (con, &kmsg) <= -1)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("Unable to route message from application.");
			}
		}
		else if (kqml_listener == con)	/* new kqml connection */
		{
			if (accept_new_connection () <= -1)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("Unable to accept new connection.");
			}
			con = -1;		/* don't look for more */
		}
		else			/* incoming message on existing connection */
		{
			if (route_msg_to_application (con, &kmsg) <= -1)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS)
					("Unable to route message to application from #%d.", con);
			}
		}

		/* If logging is on, log the kqml message */
		if (kqml_logging ())
		{
			kqml_logging_msg (kmsg);
		}
	}
}


/*
 * init_router
 * 
 * Internal data structure initialization.
 *
 */
static int
init_router ()
{
	struct sigaction act;
	sigset_t signal_mask;

	init_connections ();		     /* initialize connection table */

	sigemptyset (&signal_mask);

#	if defined (__KQML_MACH_SVR4__)
		act.sa_flags = SA_RESTART;
#	else
		act.sa_flags = 0;
#	endif

	act.sa_mask = signal_mask;
	act.sa_handler = sigpipe_handler;
	sigaction (SIGPIPE, &act, NULL); /* establish SIGPIPE handler */

	act.sa_handler = router_signal_handler;
	sigaction (SIGUSR2, &act, NULL);
	sigaction (SIGTERM, &act, NULL);
	sigaction (SIGSEGV, &act, NULL);
	sigaction (SIGBUS, &act, NULL);

	act.sa_handler = SIG_IGN;
	sigaction (SIGINT, &act, NULL);
	sigaction (SIGQUIT, &act, NULL);

	/* create an xdr stream to the app */
	setvbuf (stdin, NULL, _IOFBF, 0);
	setvbuf (stdout, NULL, _IOFBF, 0);
	if (! _kqml_init_xdr (stdin, stdout))
	{
		return -1;
	}

	/* add the application to the connections list */
	enable_connection (APP_IN, FALSE);
	connection_instream (APP_IN) = stdin;
	connection_outstream (APP_IN) = stdout;
	connection_name (APP_IN) =
		connection_host (APP_IN) = NULL;
	connection_port (APP_IN) = 0;

	return 0;
} 

/*
 * init_app
 * 
 * Initialize the connection with the application which spawned this
 * router. Communication with the application is via stdin and stdout. 
 *
 * Return value -1 on failure and 0 on success
 *
 */

static int
init_app ()
{
	kqml_message *msg;
  
#	ifndef TESTING
		/* alert the app that we are ready by sending a packet */
		msg = kqml_build_msg (KQML_HANDSHAKE, NULL, NULL, NULL, NULL);
		if (msg == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Could not build handshake message. %s.", sys_errlist[errno]);
			return -1;
		}

		if (! _kqml_send_xdr_msg (msg))
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Could not send handshake message. %s.", sys_errlist[errno]);
			return -1;
		}

		kqml_free_msg (msg);

		/* Do we need a timeout? I don't think so. */
		/* We have a sigpipe handler and this call should be interrupted
			by sigpipe if the agent dies -Chelliah */
		if (_kqml_await (APP_IN, 60) <= 0)	/* wait for a reply */
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Router unable to contact application.");
			return -1;
		}

		/* read message from application */
		msg = _kqml_fetch_xdr_msg ();
		if (msg == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Router received nothing from application.");
			return -1;
		}

		if (kqml_stricmp (kqml_get_performative (msg), KQML_HANDSHAKE) != 0)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Expected: %s, Got: %s.\n", KQML_HANDSHAKE, 
				kqml_get_performative (msg));

			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Application doesn't know the greeting.");
			return -1;
		}

		/* Get the initial agent name */
		kqml_agent_name = kqml_get_field (":content", msg, NULL, NULL);
		if (kqml_agent_name == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("No Agent name in greeting.");
			return -1;
		}

		kqml_agent_name = strdup (kqml_agent_name);
		if (kqml_agent_name == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Not enough memory for original agent.");
			return -1;
		}

		kqml_free_msg (msg);

#	else
		kqml_agent_name = "TESTING";
#	endif

	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("Router connected to Agent in Con #%d & #%d.", APP_IN, APP_OUT);

	return 0;
}

/*
 * init_listener
 * 
 * Establish a standard TCP/IP server to listen for incoming 
 * KQML connections.
 *
 * KQML selects the first available port above
 *	IPPORT_USERRESERVED+KQMLPORT_OFFSET
 * for its use, unless the application is a ans, in which case
 * it must be able to open a port exactly at  
 *	IPPORT_USERRESERVED+KQMLPORT_OFFSET
 * 
 * 
 * Returns: < 0 if failed
 *	    > 0 if succeeded
 *
 */

static int
init_listener ()
{
	int addrtype; 
	struct hostent *xhp;
	char *temp;
	char host [KQML_HOSTNAME_LEN + 1];	 	/* this host */
	int port;					   				/* port used by this agent */
  
#	ifdef __KQML_MACH_SVR4__
	if ( sysinfo (SI_HOSTNAME, host, sizeof (host)) == -1 )
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("sysinfo: %s.", sys_errlist[errno]);
		return (-1);
	}
#	else
	if (gethostname (host, sizeof (host)) <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("gethostname: %s.", sys_errlist[errno]);
		return (-1);
	}
#	endif

	if ((xhp = gethostbyname (host)) == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("gethostbyname: %s.", sys_errlist[errno]);
		return (-1);
	}

	addrtype = xhp->h_addrtype;
	if (strchr(host, '.') != NULL)
		strncpy (host, xhp->h_name, sizeof (host));
	else
	{
		struct in_addr in;

		memset (&in, 0, sizeof (in));
		memcpy (&in.s_addr, xhp->h_addr, xhp->h_length);
		strncpy (host, inet_ntoa (in), sizeof (host));
	}

	if (_kqml_im_an_ans)		/* ans must be at KQMLPORT_OFFSET */
	{
		port = IPPORT_USERRESERVED + KQMLPORT_OFFSET;
		kqml_listener = establish (port, addrtype);
		if (kqml_listener <= -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("Can't establish a ANS listener at %d. (%d)",
					port, kqml_listener);
			return -1;
		}
		else
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE)
				("Router listening in Con #%d\n", kqml_listener);
			enable_connection (kqml_listener, FALSE);
			connection_name (kqml_listener) =
				connection_host (kqml_listener) = NULL;
			connection_port (kqml_listener) = 0;
		}
	}
	else				/* establish listener above ans */
	{
		struct sockaddr_in name;
		int name_len;

		port = 0;
		/*
		int i;

		for (i = 1; i < 1000; i++)
		{
			port++;
			*/
			kqml_listener = establish (port, addrtype);
			if (kqml_listener >= 0)
			{
				_kqml_dbg_msg (KQML_DBG_VERBOSE)
					("Router listening in Con #%d\n", kqml_listener);

				enable_connection (kqml_listener, FALSE);
				connection_name (kqml_listener) =
					connection_host (kqml_listener) = NULL;
				connection_port (kqml_listener) = 0;

				name_len = sizeof (name);
				if (getsockname (kqml_listener, &name, &name_len) <= -1)
				{
					_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Can't get listener port.");
					return -1;
				}
				port = ntohs (name.sin_port);

				/*
				break;
			}
			*/
		}

		if (kqml_listener == -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Can't establish a listener at all. (%d)", kqml_listener);
			return -1;
		}
	}

	kqml_protocol = strdup ("tcp-ip");
	if (kqml_protocol == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memroy for protocol.");
		return -1;
	}

	kqml_addr_info = malloc (strlen (host) + kqml_digit_count (port) + 5);
   if (kqml_addr_info == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for addr_info.");
		return -1;
	}

	temp = kqml_addr_info;
	temp = memccpy (temp, host, 0, sizeof(host));
	*(temp - 1) = ':';
	temp += kqml_digit_to_str (port, temp);
	*temp++ = ':';
	*temp = '\0';

	return 0;
}

/*
 * establish (portnum)
 *
 * Create a server listening on a port. 
 *
 */
static int
establish (int portnum, long addrtype)
{
	struct sockaddr_in sa;
	int s;

	memset (&sa, 0, sizeof (sa));
	sa.sin_family = addrtype;
	sa.sin_port = htons (portnum);
	if ((s = socket (AF_INET, SOCK_STREAM, IPPROTO_IP)) <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("establish: socket: %s.", sys_errlist[errno]);
		return -1;
	}

	/***	This section of the code moved from after listen() to before bind()
			as otherwise there will be no effect on the socket.
					Anupama. May 20 '96.
	***/

	if (_kqml_im_an_ans)
	{
		int one;

		one = TRUE;
		if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (one)))
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("establish: setsockopt: %s.", sys_errlist[errno]);
			/* IGNORE ERRORS */
		}
	}

	if (bind (s, &sa, sizeof (sa)) < 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("establish: bind: port#%d: %s.", portnum, sys_errlist[errno]);
		return (-3);
	}
	listen (s, 5);

	dont_linger (s);

	return s;
}


/*
 * init_ans
 *
 * (Re)Establish a connection with a ans.
 *
 * Return:   0 if success
 *          -1 if error
 *
 */

static int
init_ans (char *addr_info)
{
	if (addr_info == NULL)
	{
		return -1;
	}

	/* We have to strdup 'cos connect_to_agent modifies addr_info */
	addr_info = strdup (addr_info);
	if (addr_info == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Not enough memory for ans addr.");
		return -1;
	}

	ans_connection = connect_to_agent (TRUE, kqml_ans_name, addr_info);

	free (addr_info);
	if (ans_connection <= -1)
	{
		return -1;
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Connected to ans %s:%d in #%d",
		connection_host (ans_connection), connection_port (ans_connection),
		ans_connection);

	return 0;
}

static int
register_application (int status)
{
	int res;
	kqml_message *kmsg;

	if (ans_connection == -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("register_application: Not connected to an ANS.");

		/***	Added this code because if an ANS goes down and comes up, the
				agent loses its connection with the ANS but the ANS knows of
				the existence of the agent through the persistent database.
				Here, we re-initialize the agent's connection to the ANS.
						Anupama. May 20 '96.
		***/

		if ( init_ans (getenv ("KQML_ANS")) <= -1 )
		{
			_kqml_dbg_msg(KQML_DBG_ALWAYS) 
					("register_application : Unable to reconnect to ANS.");
			return (-1);
		}
	}

	if (status == 1)
	{
		int len;
		char *buffer, *temp;

		len = strlen (kqml_addr_info) + strlen (kqml_agent_name) + 4;
		buffer = malloc (len);
		if (buffer == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("register_application: Not enough memory.");
			return -1;
		}

		temp = buffer;
		*temp++ = '(';
		temp = memccpy (temp, kqml_agent_name, 0, len);
		*(temp - 1) = ' ';
		temp = memccpy (temp, kqml_addr_info, 0, len);
		*(temp - 1) = ')';
		*temp = '\0';

		kmsg = kqml_build_msg ("register", buffer, kqml_ans_name, "t", NULL);
		free (buffer);
	}
	else
	{
		kmsg = kqml_build_msg ("unregister", kqml_agent_name, kqml_ans_name, 
			NULL, NULL);
	}

	if (kmsg == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("register/unregister: Could not build msg.");
		return -1;
	}
	res = send_parsed_msg (kmsg, ans_connection, 
		connection_outstream (ans_connection));
	kqml_free_msg (kmsg);

	if (res <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("%s: fputs: %s.",
			(status == 1 ? "register" : "unregister"), sys_errlist[errno]);
		return -1;
	}

	return 0;
}

static int
confirm_registration ()
{
	kqml_message *kmsg;
	int res;
	char *performative, *content;

	/* Wait for a reply */ 
	res = _kqml_await (ans_connection, 30);
	if (res <= 0)
	{
		_kqml_dbg_msg(KQML_DBG_ALWAYS) 
			("ANS timeout error. %s.", sys_errlist[errno]);
		return FALSE;
	}

	kmsg = parse_kqml (ans_connection);
	if (! kmsg)			/* EOF or error encountered */
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("ANS is gone.");
		return FALSE;
	}

	if (kmsg == PARSE_ERROR)	/* read error */
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Unexpected message from ANS.");
		return FALSE;
	}
		
	performative = kqml_get_performative (kmsg);
	content = kqml_get_field (":content", kmsg, NULL, NULL);
	if (kqml_stricmp (performative, "reply") != 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Received a %s from ANS with [%s].", performative, content);
		kqml_free_msg (kmsg);
		return FALSE;
	}

	free (kqml_agent_name);
	content = strchr (content, '(');
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Received an illegal reply from the ANS.");
		kqml_free_msg (kmsg);
		return FALSE;
	}
	content++;

	kqml_agent_name = strtok (content, " \t\n");
	kqml_ans_name = strtok (NULL, " \t\n)");

	if (kqml_agent_name)
	{
		kqml_agent_name = strdup (kqml_agent_name);
	}
	if (kqml_ans_name)
	{
		kqml_ans_name = strdup (kqml_ans_name);
	}
	kqml_free_msg (kmsg);

	if (kqml_agent_name == NULL || kqml_ans_name == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Not enough memory to store agent name or ans name.");
		return FALSE;
	}

	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Registered agent [%s] at [%s] with [%s]", 
		kqml_agent_name, kqml_addr_info, kqml_ans_name);

	return TRUE;
}

static int
complete_handshake ()
{
	int reply_len;
	char *reply_str, *temp;
	kqml_message *greeting;

	reply_len = strlen (kqml_agent_name) + strlen (kqml_ans_name) +
		strlen (kqml_protocol) + strlen (kqml_addr_info) + 6;
	reply_str = malloc (reply_len);
	if (reply_str == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Router could not construct handshake string.");
		return FALSE;
	}

	temp = reply_str;
	*temp++ = '(';
	temp = memccpy (temp, kqml_agent_name, 0, reply_len);
	*(temp - 1) = ' ';
	temp = memccpy (temp, kqml_ans_name, 0, reply_len);
	*(temp - 1) = ' ';
	temp = memccpy (temp, kqml_protocol, 0, reply_len);
	*(temp - 1) = ' ';
	temp = memccpy (temp, kqml_addr_info, 0, reply_len);
	*(temp - 1) = ')';
	*temp = '\0';

	greeting = kqml_build_msg (KQML_HANDSHAKE, reply_str, NULL, NULL, NULL);
	free (reply_str);
	if (greeting == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Router could not construct greeting message to app.");
		return FALSE;
	} 
	if (! _kqml_send_xdr_msg (greeting)) 
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Router could not send greeting message to agent.");
		return FALSE;
	}

	kqml_free_msg (greeting);

	return TRUE;
}


static void
sigpipe_handler (int sig, void *dummy1, void *dummy2)
{
	_kqml_dbg_msg (KQML_DBG_ALWAYS) ("received a SIGPIPE."); 
	return;
}

static void
router_signal_handler (int signo, void *dummy1, void *dummy2)
{
	if (signo == SIGUSR2)
		dump_connection_table ();
	else
		router_exit (signo);
}

static void
router_exit (int ret_code)
{
	static int going = FALSE;

	if (! going)
		going = TRUE;
	else
		return;

	if (! _kqml_root_ans)
	{
		/* unregister the application */
		register_application (FALSE);
	}

	/* exit */
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Exiting. %d.", ret_code);

	exit (ret_code);   
}


static int
kqml_logging_msg (kqml_message *kmsg)
{
	kqml_message *log_kmsg;
	char *content;
	int  value_len;

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Here in kqml_logging_msg");
	if (kmsg == NULL)
	{
		return -1;
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("check the logger connection");
	if (logger_connection == -1)
	{
		if ((logger_connection = 
			get_connection_to_remote (KQML_LOGGER_NAME, NULL)) < 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE)
				("Unable to reach KQML-LOGGER agent");
			return -1;
		}
	}
	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml_logging_msg: made connection");

	/* create a printed representation of the KQML message */
	content = kqml_sprint_msg (NULL, &value_len, kmsg);
	_kqml_dbg_msg (KQML_DBG_VERBOSE)
		("kqml_logging_msg: content=%.*s\n", value_len-1, content);

	/* Create the message to be sent to the logging agent */
	log_kmsg = kqml_build_msg ("tell", NULL, KQML_LOGGER_NAME, NULL,
		":sender", kqml_agent_name, NULL);
	kqml_put_field (":content", log_kmsg, content, value_len - 1);
	_kqml_dbg_kmsg (KQML_DBG_VERBOSE) (log_kmsg);

	if (send_parsed_msg (log_kmsg, logger_connection,
		connection_outstream (logger_connection)) < 0)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE)
			("Failed to send message to KQML-LOGGER agent");
		return -2;
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Done with kqml_logging_msg");
	return 0;
}


/*
 * Find a system by symbolic name and return an fd
 *
 */
static int
get_connection_to_remote (char *symbolic_name, char *lookup_info)
{
	int i, lookup_ans_connection;
	char *lookup_ans_name;

	if (symbolic_name == NULL)
	{
		_kqml_dbg_msg(KQML_DBG_ALWAYS) 
			("Unable to find :receiver field in message.");
		return (-1);
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Searching agent connection [%s].", 
		symbolic_name);

	/* check the local connection table */
	for (i = 0; i < MAX_ROUTER_CONNECTIONS ; i++)
	{
		if (connection_create_time (i) != 0 && 
			connection_name (i) != NULL &&
			kqml_stricmp (symbolic_name, connection_name (i)) == 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("Known agent [%s] at %s:%d in Con #%d.", connection_name (i), 
					connection_host (i), connection_port (i), i);
			return (i);
		}
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Looking up agent [%s].", symbolic_name);

	if (_kqml_im_an_ans)
	{
		/* If we are an ANS */
		char *protocol, *temp, *addr_info;

		if (lookup_info == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("No :kqml-address info.");
			return -1;
		}

		temp = strchr (lookup_info, '(');
		if (temp == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				(":kqml-address format error [%s]", lookup_info);
			return -1;
		}
		temp++;

		lookup_ans_name = strtok (temp, " \t\n");
		if (lookup_ans_name == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				(":kqml-address format (name) error [%s]", lookup_info);
			return -1;
		}

		/* Ignore the protocol */
		protocol = strtok (NULL, " \t\n");
		if (protocol == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				(":kqml-address format (protocol) error [%s]", lookup_info);
			return -1;
		}
	
		addr_info = strtok (NULL, " \t\n)");
		if (addr_info == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				(":kqml-address format (addr_info) error [%s]", lookup_info);
			return -1;
		}

		lookup_ans_connection = connect_to_agent (FALSE, lookup_ans_name, 
			addr_info);

		if (lookup_ans_connection == -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("Could not connect to %s", lookup_ans_name);
			return -1;
		}

		if (kqml_stricmp (symbolic_name, lookup_ans_name) == 0)
		{
			return lookup_ans_connection;
		}
	}
	else
	{
		/* else ask our ANS */
		if (ans_connection <= -1)
		{
			if ( init_ans (getenv ("KQML_ANS")) <= -1 )
			{
				_kqml_dbg_msg(KQML_DBG_ALWAYS) ("Unable to reconnect to ANS.");
				_kqml_dbg_msg(KQML_DBG_ALWAYS) 
					("Unable to route to %s.", symbolic_name);
				return (-1);
			}
		}

		lookup_ans_connection = ans_connection;
		lookup_ans_name = kqml_ans_name;
	}
  
	return lookup_and_connect (symbolic_name, lookup_ans_name, 
		lookup_ans_connection);
}

/*
 * Lookup an agent's address with an ANS and connect to the agent.
 *
 * Only one kind of message is currently expected from the ANS.
 * 
 * (reply :content "name:hostname:port")
 *
 */

static int
lookup_and_connect (char *agent_name, char *lookup_ans_name, int lookup_ans_con)
{
	kqml_message *kmsg;
	int res;
	char *content, *temp;

	kmsg = kqml_build_msg ("ask-one", agent_name, lookup_ans_name, "t", 
		":sender", kqml_agent_name, ":ontology", kqml_protocol, NULL);
	if (! kmsg)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 	
			("lookup_and_connect: could not build lookup msg.");
		return -1;
	}

	res = send_parsed_msg (kmsg, lookup_ans_con, 
		connection_outstream (lookup_ans_con));
	kqml_free_msg (kmsg);

	if (res <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("lookup_and_connect: error on write to ANS %s.", lookup_ans_name);
		return -1;
	}

	/* Wait for a reply */
	res = _kqml_await (lookup_ans_con, 30);

	if (res <= 0)
	{
		_kqml_dbg_msg(KQML_DBG_ALWAYS) 
			("ANS timeout error. %s.", sys_errlist[errno]);
		return -1;
	}

	kmsg = parse_kqml (lookup_ans_con);
	if (! kmsg)			/* EOF or error encountered */
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Lookup ANS %s is gone.", 
			lookup_ans_name);
		return -1;
	}
	else if (kmsg == PARSE_ERROR)	/* read error */
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Unexpected message from lookup ANS %s.", lookup_ans_name);
		return -1;
	}

	/* check message */
	temp = kqml_get_performative (kmsg);
	if (kqml_stricmp (temp, "reply") != 0)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("Received a %s from the lookup ANS.", temp);
		return -1;
	}

	content = kqml_get_field (":content", kmsg, NULL, NULL);
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("No content in message from the lookup ANS.");
		return -1;
	}
  
	temp = strchr(content, ':');
	if (temp == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Couldn't parse agent name from message: [%s]", content);
		return -1;
	}
	content = temp + 1;

	res = connect_to_agent (FALSE, agent_name, content);

	kqml_free_msg (kmsg);

	return res;
}

static int
connect_to_agent (int env_flag, char *agent_name, char *addr_info)
{
	char *host_name;
	int host_port, agent_con;

	/* NOTE: agent_name can be NULL */
	/* (IMS) but we give it to printf anyway... doh */
	/* better to fix! */
	if (agent_name == NULL)
	{
	    agent_name = strdup("UNNAMED");
	    if (agent_name == NULL) /* failed malloc in strdup() */
	    {
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
		    ("Failed malloc in connect_to_agent()");
		exit(-1);
	    }
	}
	parse_tcp_info (env_flag, addr_info, &host_name, &host_port);
	if (host_name == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Couldn't parse host name from addr_info : [%s]", addr_info);
		return -1;
	}

	agent_con = _kqml_tcp_to_service (host_port, host_name);
	agent_con = verify_connection_range (agent_con);

	if (agent_con <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Couldn't connect with %s at %s:%d.",
			agent_name, host_name, host_port);
	}      
	else
	{
		dont_linger (agent_con);

		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("Connected to agent [%s] at %s:%d in Con #%d.", agent_name,
			host_name, host_port, agent_con);

		if (agent_name)
		{
			connection_name (agent_con) = strdup (agent_name);
			if (connection_name (agent_con) == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("Could not save agent %s's name for Con #%d, at %s:%d.", 
						agent_name, agent_con, host_name, host_port);
			}
		}

		connection_host (agent_con) = strdup (host_name);
		if (connection_host (agent_con) == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("Could not save agent %s's host for Con #%d, at %s:%d.",
					agent_name, agent_con, host_name, host_port);
		}

		connection_port (agent_con) = host_port;

		enable_connection (agent_con, TRUE);
	}
	return agent_con;
}

/* parse_tcp_info
 * Break up a ans:port specification.
 *
 */
static int
parse_tcp_info (int env_flag, char *param, char **host, int *port)
{
	char *p;
  
	while (*param == ' ') 
		param++; /* skip leading white space */

	*host = param;
	p = strchr (param, ':');	 /* find end of host name */
	if (p == NULL) 
	{
		if (env_flag && *param)
		{
			/* We can just have the host name */
			*host = param;
			return TRUE;
		}
		else
		{
			*host = NULL;
			return FALSE;
		}
	}

	*p = '\0';			 /* terminate the name */
	*port = atoi (p+1);
	if (*port == 0)
	{
		if (env_flag)
		{
			/* Assume default port */
			*port = IPPORT_USERRESERVED + KQMLPORT_OFFSET;
		}
		else
		{
			*host = NULL;
			return 0;
		}
	}

	return 1;
}


static kqml_message *
parse_kqml (int con)
{
	kqml_message *kmsg = NULL;
	int ret_value;
  

	if (! connection_instream (con))
	{
		/* When we created the connection, we created with streams enabled. 
			But we do not have a stream. So, most probably fdopen failed
			due to too many open streams */
		/* So use kqml_parse_fd */
		
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("About to call kqml_parse_fd.");
		ret_value = kqml_parse_fd (con, &kmsg);
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Returned from kqml_parse_fd.");
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("About to call kqml_parse_stream.");
		ret_value = kqml_parse_stream (connection_instream (con), &kmsg);
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Returned from kqml_parse_stream.");
	}


	if (ret_value == KQML_PARSE_EOF || ret_value == KQML_PARSE_TIMEOUT)
	{
		_kqml_dbg_msg (KQML_DBG_WARNING) 
			("%s connection. Con #%d to agent %s at %s:%d.",
				(ret_value == KQML_PARSE_EOF ? "Broken" : "Timed out"),
				con, 
				(connection_name (con) ? connection_name (con) : "UNKNOWN"),
				(connection_host (con) ? connection_host (con) : "UNKNOWN"),
				connection_port (con));
		close (con);
		disable_connection (con);
		if (con == ans_connection)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("ANS is gone.");
			ans_connection = -1;
		}
		else if (con == logger_connection)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Logger is gone.");
			logger_connection = -1;
		}
		return NULL;
	}
	else if (ret_value != KQML_PARSE_OK)
	{
		return PARSE_ERROR;
	}

	return kmsg;
}


/*
 * Accept a new connection on the KQML listener port
 *
 */
static int
accept_new_connection ()
{
	struct sockaddr_in	isa;	/* address of socket */
	int con;							/* new connection */
	struct hostent *hp;
	int sz, retry_count;
	char *host_address;

	sz = sizeof (isa);

	retry_count = 5;
	while (retry_count--)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("Calling accept new connection in Con #%d (Attempt %d).", 
			kqml_listener, 5 - retry_count);

		con = accept (kqml_listener, &isa, &sz);
		if (con <= -1)
		{
			_kqml_dbg_msg (KQML_DBG_WARNING) 
				("accept_new_connection: %s.", sys_errlist[errno]);

			if (errno == EMFILE)	/* too many open files */
			{
				if (close_nonpending_connection () <= -1)
				{
					_kqml_dbg_msg (KQML_DBG_ALWAYS) 
						("Can't close any connections.");
					break;	/* No point in retrying */
				}
			}
		}
		else
		{
			con = verify_connection_range (con);
			if (con == -1)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) ("Reached max connections.");
				break; /* No point in retrying */
			}
			else
			{
				break;
			}
		}
	}

	if (con != -1)
	{
		dont_linger (con);

		/* save host and socket info */
		hp = gethostbyaddr ((char *)&isa.sin_addr, sizeof (isa.sin_addr), 
			isa.sin_family);
		if (hp)
		{
			host_address = hp->h_name;
		}
		else
		{
			host_address = inet_ntoa (isa.sin_addr);
			_kqml_dbg_msg (KQML_DBG_WARNING) 
				("Couldn't get hostname for %s", host_address);
		}

		connection_port (con) = ntohs (isa.sin_port);

		connection_host (con) = strdup (host_address);
		if (connection_host (con) == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Could not save host for Con #%d at %s:%d.",
				con, host_address, connection_port (con));
		}

		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("Connection from host %s:%d in Con #%d.",
			(connection_host (con) ? connection_host (con) : "UNKNOWN"), 
			connection_port (con), con);

		enable_connection (con, TRUE);
		connection_name (con) = NULL;
	}

	return con;
}

static void
dont_linger (int con)
{
	struct linger linger;

	linger.l_onoff = TRUE;
	linger.l_linger = 0;

	/* When a socket is closed, if there is data to be delivered,
		the system lingers a while trying to deliver that data.
		This can keep the port occupied for sometime. Reduce that time
		by eliminating the linger time.
		Since, this is just the listener, it will never be used to
		send out data and we can use a linger time of 0
	*/
	if (setsockopt (con, SOL_SOCKET, SO_LINGER, &linger, sizeof (linger)))
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("dont_linger: setsockopt: %s.", sys_errlist[errno]);
		/* IGNORE ERRORS */
	}
}

/* 
 * KQML keeps connections to other agents open as much as
 * possible. But as it approaches the maximum number of allowable open
 * sockets it must start to close some in order to open new ones. This
 * routine selects a least recently used socket for which there is no
 * reply message expected. That is, we avoid closing sockets on which
 * we have received message with :reply-with = t but which we haven't
 * replied to yet.) 
 */
static int
close_nonpending_connection ()
{
	int con = -1;
	int i;
	unsigned long t = ULONG_MAX;

	/* look for the oldest non-pending connection (but not stdin or stdout!) */
	for (i = 2; i < MAX_ROUTER_CONNECTIONS; i++)
	{
		/* Don't use any of the connections below */
		if ((i == ans_connection) || (i == kqml_listener) ||
			(i == logger_connection))
		{
			continue;
		}
		if (! connection_pending (i))
		{
			if (connection_last_use_time (i) < t)
			{
				t = connection_last_use_time (i);
				con = i;
			}
		}
	}

	_kqml_dbg_msg (KQML_DBG_WARNING) 
		("Closed non-pending Con #%d to agent %s at %s:%d.", 
			con,
			(connection_name (con) ? connection_name (con) : "UNKNOWN"),
			(connection_host (con) ? connection_host (con) : "UNKNOWN"),
			connection_port (con));
	close (con);
	disable_connection (con);

	return (con);
}

static int
verify_connection_range (int con)
{
	int new_con = -1;

	if (! connection_valid (con))
	{
		if (con != -1)
		{
			new_con = close_nonpending_connection ();
			if (new_con != -1)
			{
				dup2 (con, new_con);
			}
			close (con);
		}
	}
	else
	{
		new_con = con;
	}

	return new_con;
}


/*
 * Find or wait for the next socket on which there is data.
 *
 *
 */
static int
get_active_connection (int previous)
{
	int i, con, size;
	fd_set empty_set;
	static int dtablesize = -1;
	static fd_set active_servers;
  
	if (dtablesize == -1)
	{
		FD_ZERO (&active_servers);
#		if defined (__KQML_MACH_SVR4__)
			dtablesize = ulimit (UL_GDESLIM, 0);
#		else
			dtablesize = ulimit (4, 0);
#		endif
	}

	/* First check for any remaining data in the stream */
	/* buffer of the previously returned socket */
	/* If that connection is still valid */
	if (previous >= 0 && connection_create_time (previous) != 0)
	{
		if (previous != APP_IN)
			size = _kqml_pending_bytes (previous, connection_instream (previous));
		else
			size = _kqml_msgs_in_queue ();

		if (size <= -1)
			return -1;

		if (size != 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("Returning to previous Con #%d -- %d", previous, size);
			return (previous);	                 /* keep working on this one */
		}
  	}

	FD_ZERO (&empty_set);
	while (1)
	{
		/* Next consult the set of active connections from */
      /* the previous call to get_active_connection. */
		for (con = 0; con < MAX_ROUTER_CONNECTIONS; con++)
		{
			if (FD_ISSET (con, &active_servers))
			{
				if (con != APP_IN)
					size = _kqml_pending_bytes (con, connection_instream (con));
				else
					size = _kqml_msgs_in_queue ();
				_kqml_dbg_msg (KQML_DBG_VERBOSE) 
					("Con #%d is active -- %d", con, size);
				FD_CLR (con, &active_servers);
				return con;
			}
		}

		/* look for a new active set */
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Looking for a new active con.");
		do
		{
			active_servers = servers;
			i = select (dtablesize, &active_servers, NULL, NULL, NULL);
		}
		while (( i < 0) && (errno == EINTR));

		if (i < 0)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS) 
				("Select error: %s", sys_errlist[errno]);
			FD_ZERO (&active_servers);
			return (-1);
		}
	}
}


/*
 * When messages come in for the application they are parsed and then
 * xdr-ed to the application.
 *
 */   
static int
route_msg_to_application (int con, kqml_message **k_msg)
{
	kqml_message *kmsg;
	struct timeval	beg_kmsg_read, end_kmsg_read;
	int ret_value;

	/* update connection */
	connection_last_use_time (con) = time (NULL);

	/* read and copy message */

	*k_msg = NULL;
	gettimeofday (&beg_kmsg_read, NULL);
	kmsg = parse_kqml (con);
	gettimeofday (&end_kmsg_read, NULL);

	if (kmsg == PARSE_ERROR)
	{
		_kqml_dbg_msg (KQML_DBG_WARNING) 
			("Read from Con #%d failed with parse error.", con);
		return -1;
	}
	else if (kmsg == NULL)
	{
		return 0;
	}
	*k_msg = kmsg;

	ret_value = send_to_application (con, kmsg);

	if (ret_value != -1)
	{
		set_kmsg_time (kqml_get_field (KMSG_ID_KEYWORD, kmsg, NULL, NULL), 
			KMSG_READ_TIME, &beg_kmsg_read, &end_kmsg_read);
	}

	return ret_value;
}

static int
send_to_application (int con, kqml_message *kmsg)
{
	char *kmsg_id, *tag, *perf;
	int ret_value;

	kmsg->connection = con;
	perf = kqml_get_performative (kmsg);
	tag = kqml_get_field (":in-reply-to", kmsg, NULL, NULL);

	if (kqml_stricmp("kqml-admin", perf) == 0)
	{
		ret_value = handle_admin_msg (kmsg);
		if (ret_value != 1)
		{
			return ret_value;
		}
	}

	if (!_kqml_send_xdr_msg (kmsg))
	{
		_kqml_dbg_msg (KQML_DBG_WARNING) ("Unable to write to application.");
		return -1;
	}
	kill (agent_pid, SIGIO);

	kmsg_id = kqml_get_field (KMSG_ID_KEYWORD, kmsg, NULL, NULL);
	if ( kmsg_id == NULL )
	{
		kmsg_id = "unknown";
		kqml_put_field (KMSG_ID_KEYWORD, kmsg, kmsg_id, -1);
	}

	connection_pending (con) = !(_kqml_get_reply_with_field (kmsg) == NULL);

	return 0;
}

static int
route_msg_from_application (int con, kqml_message **k_msg)
{
	char *receiver;
	kqml_message *kmsg;
	struct timeval beg_kmsg_write, end_kmsg_write;
	int ret_value;
  
	/* read the message */
	*k_msg = NULL;
	kmsg = _kqml_fetch_xdr_msg ();
	if (!kmsg)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS)
			("The application appears to have died. Exiting.");
		router_exit (1);
	}
	*k_msg = kmsg;

	receiver = kqml_get_field (":receiver", kmsg, NULL, NULL);
	if (receiver != NULL && kqml_stricmp (receiver, kqml_agent_name) == 0)
	{
		return send_to_application (con, kmsg);
	}

	ret_value = send_to_remote (con, kmsg, &beg_kmsg_write, &end_kmsg_write);

	if (ret_value != -1)
	{
		set_kmsg_time (kqml_get_field (KMSG_ID_KEYWORD, kmsg, NULL, NULL), 
			KMSG_WRITE_TIME, &beg_kmsg_write, &end_kmsg_write);
	}

	return ret_value;
}

static int
send_to_remote (int con, kqml_message *kmsg, struct timeval *beg, 
	struct timeval *end)
{
	char *receiver, *lookup_info;

	lookup_info = kqml_get_field (":kqml-address", kmsg, NULL, NULL);

	/* determine the destination */
	if (kmsg->connection == 0)
	{
		receiver = kqml_get_field (":receiver", kmsg, NULL, NULL);
		kmsg->connection = get_connection_to_remote (receiver, lookup_info);

		if (kmsg->connection <= 0)
		{
			send_error_reply 
				("Unable to find a destination for outbound message.", kmsg);
			return -1;
		}
	}

	if (lookup_info)
	{
		/* Remove the address field */
		kqml_put_field (":kqml-address", kmsg, NULL, 0);
	}

	/* send it */
	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("Router sending to Con #%d\n", kmsg->connection);
	_kqml_dbg_kmsg (KQML_DBG_VERBOSE) (kmsg);

	gettimeofday (beg, NULL);
	if (send_parsed_msg (kmsg, kmsg->connection, 
		connection_outstream (kmsg->connection)) < 0)
	{
		_kqml_dbg_msg (KQML_DBG_INFORM) 
			("Unable to send parsed message to Con #%d", kmsg->connection);
		send_error_reply ("Unable to send outbound message.", kmsg);
		return -1;
	}
	gettimeofday (end, NULL);

	return 0;
}


/*
 * send a message to the application indicating that the last message received
 * could not be transmitted.
 * 
 */
static void
send_error_reply (char *msg, kqml_message *oldmsg)
{
	kqml_message *kmsg;
	char *reply_with;

	reply_with = _kqml_get_reply_with_field (oldmsg);
	/* send reply back to the application only if it is expecting a 
		reply */
	if (reply_with)
	{
		char *sender;

		sender = kqml_get_field (":sender", oldmsg, NULL, NULL);

		/* build the message */
		kmsg = kqml_build_msg ("error", msg, sender, reply_with, NULL);
	
		if (kmsg)
		{
			/* do not check for error of this kqml_put_field */
			kqml_put_field (":sender", kmsg, sender, -1);

			kqml_put_field (":in-reply-to", kmsg, reply_with, -1);

			/* send it to the application */
			if (!_kqml_send_xdr_msg (kmsg))
			{
				_kqml_dbg_msg (KQML_DBG_WARNING) 
					("send_error_reply: Unable to write to appln.");
			}
			kill (agent_pid, SIGIO);
			kqml_free_msg (kmsg);
		}
		else
		{
			_kqml_dbg_msg (KQML_DBG_WARNING)
				("send_erro_reply: Unable to build message.");
		}
	}
	return;
}

/* Returns -1 on error, 0 on success and 1 if the agent has to handle it */
static int 
handle_admin_msg (kqml_message *kmsg)
{
	char *content, *content_copy;

	content = kqml_get_field(":content", kmsg, NULL, NULL);
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Missing content field");
		return -1;
	}

	/* 
	 * We may decide that the agent has to pass this message. If so,
	 * the content field should be intact. 
	 */
	content_copy = content = strdup (content);
	if (content_copy == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Could not copy content");
		return 1;
	}

	content = strchr (content, '(');
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Content format error");
		free (content_copy);
		return -1;
	}
	content++;

	content = strtok (content, " \t\n)");
	if (content == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Empty content field");
		free (content_copy);
		return -1;
	}

	if (kqml_stricmp (content, "timings") == 0)
	{
		content = strtok (NULL, " \t\n)");
		if (content == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Empty timings command");
			free (content_copy);
			return -1;
		}

		if (kqml_stricmp (content, "on") == 0)
		{
			kqml_timings_start ();
			return 0;
		}

		if ( kqml_stricmp(content, "off") == 0)
		{
			kqml_timings_stop ();
			return 0;
		}

		free (content_copy);
		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("kqml-admin: Illegal timings command [%s]", content);
		return -1;
	}

	if (kqml_stricmp (content, "logging") == 0)
	{
		content = strtok (NULL, " \t\n)");
		if (content == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("kqml-admin: Empty logging command");
			free (content_copy);
			return -1;
		}

		if ( kqml_stricmp(content, "on") == 0)
		{
			kqml_logging_start ();
			return 0;
		}

		if ( kqml_stricmp(content, "off") == 0)
		{
			kqml_logging_stop ();
			return 0;
		}

		free (content_copy);
		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("kqml-admin: Illegal logging command [%s]", content);
		return -1;
	}

	if (kqml_stricmp(content, "name") == 0)
	{
		char *sender_fac, *reply_with;
		kqml_message *msg;

		sender_fac = kqml_get_field (":sender", kmsg, NULL, NULL);
		reply_with = kqml_get_field (":reply-with", kmsg, NULL, NULL);
		if (sender_fac == NULL || reply_with == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("kqml-admin: No sender/reply-with specified in name request");
			free (content_copy);
			return -1;
		}

		/*	Check to see if the ANS who is sending this message is the
			right one	*/
		if ( kqml_stricmp (kqml_ans_name, sender_fac) != 0 )
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("kqml-admin: name request from a non-ANS. Ignored.");
			free (content_copy);
			return 0;
		}

		msg = kqml_build_msg ("reply", kqml_agent_name, sender_fac, NULL,
			":sender", kqml_agent_name, ":in-reply-to", reply_with, NULL);

		if (msg == NULL)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("kqml-admin: name response failed. could not build message");
			free (content_copy);
			return -1;
		}

		if (send_parsed_msg (msg, kmsg->connection, 
			connection_outstream (kmsg->connection)) <= -1)
		{
			kqml_free_msg (msg);
			_kqml_dbg_msg (KQML_DBG_VERBOSE) 
				("kqml-admin: name response failed. could not send message");
			free (content_copy);
			return -1;
		}
		free (content_copy);
		kqml_free_msg (msg);
	}

	if (kqml_stricmp (content, "shutdown") == 0)
	{
		free (content_copy);
		return 1;
	}
	else
	{
		_kqml_dbg_msg (KQML_DBG_VERBOSE)
			("kqml-admin: Illegal command [%s]. Ignored.", content);
		free (content_copy);
		return -1;
	}
}
